#include <CUnit/CUnit.h>
#include <CUnit/Basic.h>

#include <stdio.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>

static int empty_init(void)
{
    return 0;
}

static int empty_cleanup(void)
{
    return 0;
}

/********************************************************************
 * initng_toolbox.c tests
 ********************************************************************/
#include "../src/initng_toolbox.h"

/* 
 * invoke initng_calloc()
 */
static void test_initng_calloc(void)
{
    char *s;

    s = initng_calloc(1, 101);
    CU_ASSERT(s != NULL);
    free(s);
}

/* 
 * invoke initng_realloc() to both grow and shrink a block
 */
static void test_initng_realloc(void)
{
    char *s;

    s = initng_calloc(1, 100);
    CU_ASSERT(s != NULL);
    s = initng_realloc(s, 50);
    CU_ASSERT(s != NULL);
    s = initng_realloc(s, 200);
    CU_ASSERT(s != NULL);
    free(s);
}

/*
 * invoke i_strdup() and verify that it works
 */
static void test_i_strdup(void)
{
    char *p;
    char *s;

    p = "the quick brown fox jumps over the lazy dog";
    s = i_strdup(p);
    CU_ASSERT(s != NULL);
    CU_ASSERT(strcmp(s, p) == 0);
    free(s);
}

/*
 * invoke i_strndup() and verify that it works
 */
static void test_i_strndup(void)
{
    char *p;
    char *s;

    p = "the quick brown fox jumps over the lazy dog";
    s = i_strndup(p, 100);
    CU_ASSERT(s != NULL);
    CU_ASSERT(strcmp(s, p) == 0);
    free(s);

    s = i_strndup(p, 3);
    CU_ASSERT(s != NULL);
    CU_ASSERT(strcmp(s, "the") == 0);
    free(s);
}

/* 
 * We want to test the memory functions when they handles the out of 
 * memory condition.
 *
 * 1. reduce the maximum amount of memory usable for the process, so
 *    that the alloc fails
 * 2. set an alarm() and a signal handler that will increase the
 *    maximum amount of memory back to the original level, so that
 *    the function eventually gets memory and returns
 * 3. invoke function
 */

static struct rlimit rlim_save;

static void test_out_of_memory_remove_limit(int signum)
{
    if (setrlimit(RLIMIT_AS, &rlim_save) != 0)
    {
        fprintf(stderr, "%s:%d Error %d with setrlimit(); %s\n",
                __FILE__, __LINE__, errno, strerror(errno));
        exit(1);
    }
}

static void test_out_of_memory(void)
{
    static struct rlimit rlim;
    struct sigaction sa;
    char *s;
    char *p;
    char *q;
    struct timeval start;
    struct timeval end;
    int usec;

    /* get the current memory limit */
    if (getrlimit(RLIMIT_AS, &rlim_save) != 0)
    {
        fprintf(stderr, "%s:%d Error %d with getrlimit(); %s\n",
                __FILE__, __LINE__, errno, strerror(errno));
        exit(1);
    }

    /* set up the signal handler for the alarm signal */
    memset(&sa, 0, sizeof(struct sigaction));
    sa.sa_handler = test_out_of_memory_remove_limit;
    sa.sa_flags = SA_RESTART;
    if (sigaction(SIGALRM, &sa, NULL) != 0)
    {
        fprintf(stderr, "%s:%d Error %d with sigaction(); %s\n",
                __FILE__, __LINE__, errno, strerror(errno));
        exit(1);
    }

    /* set the memory limit to a very low value */
    rlim = rlim_save;
    rlim.rlim_cur = 1;
    if (setrlimit(RLIMIT_AS, &rlim) != 0)
    {
        fprintf(stderr, "%s:%d Error %d with setrlimit(); %s\n",
                __FILE__, __LINE__, errno, strerror(errno));
        exit(1);
    }

    /** initng_calloc **/

    /* set an alarm, and then invoke calloc */
    gettimeofday(&start, NULL);
    alarm(1);
    s = initng_calloc(1, 1024 * 1024);
    gettimeofday(&end, NULL);

    /* verify allocation worked */
    CU_ASSERT(s != NULL);

    /* make sure it actually took time and restarted */
    usec = (end.tv_sec - start.tv_sec) * 1000000;
    usec += end.tv_usec - start.tv_usec;
    CU_ASSERT(usec > 1000000);

    /** initng_realloc **/

    /* set our limit again */
    if (setrlimit(RLIMIT_AS, &rlim) != 0)
    {
        fprintf(stderr, "%s:%d Error %d with setrlimit(); %s\n",
                __FILE__, __LINE__, errno, strerror(errno));
        exit(1);
    }

    /* and try with realloc */
    gettimeofday(&start, NULL);
    alarm(1);
    s = initng_realloc(s, 2 * 1024 * 1024);
    gettimeofday(&end, NULL);

    /* verify resize worked */
    CU_ASSERT(s != NULL);

    /* make sure it actually took time and restarted */
    usec = (end.tv_sec - start.tv_sec) * 1000000;
    usec += end.tv_usec - start.tv_usec;
    CU_ASSERT(usec > 1000000);

    /** i_strdup **/

    /* set our limit again */
    if (setrlimit(RLIMIT_AS, &rlim) != 0)
    {
        fprintf(stderr, "%s:%d Error %d with setrlimit(); %s\n",
                __FILE__, __LINE__, errno, strerror(errno));
        exit(1);
    }

    /* and try with strdup */
    gettimeofday(&start, NULL);
    alarm(1);
    memset(s, 'x', 1024 * 1024);
    s[1024 * 1024] = '\0';
    p = i_strdup(s);
    gettimeofday(&end, NULL);

    /* verify strdup worked */
    CU_ASSERT(p != NULL);

    /* make sure it actually took time and restarted */
    usec = (end.tv_sec - start.tv_sec) * 1000000;
    usec += end.tv_usec - start.tv_usec;
    CU_ASSERT(usec > 1000000);

    /** i_strndup **/

    /* set our limit again */
    if (setrlimit(RLIMIT_AS, &rlim) != 0)
    {
        fprintf(stderr, "%s:%d Error %d with setrlimit(); %s\n",
                __FILE__, __LINE__, errno, strerror(errno));
        exit(1);
    }

    /* and try with strdup */
    gettimeofday(&start, NULL);
    alarm(1);
    q = i_strndup(s, 1024 * 512);
    gettimeofday(&end, NULL);

    /* verify strndup worked */
    CU_ASSERT(q != NULL);

    /* make sure it actually took time and restarted */
    usec = (end.tv_sec - start.tv_sec) * 1000000;
    usec += end.tv_usec - start.tv_usec;
    CU_ASSERT(usec > 1000000);

    /* free memory */
    free(s);
    free(p);
    free(q);
}

/*
 * invoke set_proc_title() and verify that it works
 *
 * In order to do this, we use "ps" to find our own process id in the
 * process list, and look at the name there. This test depends on 
 * "ps" being available. It would probably be better to look at 
 * /proc/$pid, but this is easy enough.
 *
 * Note that this is not a very good test, as it depends on the amount
 * of space available for the process to set its own title. This is 
 * the amount of space allocated by the kernel for argv[] and env[],
 * which might be too little for the actual initng process.
 */

static char *get_process_name(void)
{
    FILE *fp;
    int pid;
    char line[1024];
    char *s;
    char *name;

    fp = popen("ps wwwwo pid,command", "r");
    if (fp == NULL)
    {
        fprintf(stderr, "%s:%d Error with popen()\n", __FILE__, __LINE__);
        exit(1);
    }
    pid = getpid();
    s = NULL;
    while (fgets(line, sizeof(line), fp) != NULL)
    {
        if (atoi(line) == pid)
        {
            s = line;
            while (isspace(*s))
                s++;                        /* skip leading whitespace */
            while (isdigit(*s))
                s++;                        /* skip pid */
            s++;                            /* skip space after pid */

            /* copy the name */
            name = strdup(s);
            if (name == NULL)
            {
                fprintf(stderr, "%s:%d Out of memory\n", __FILE__, __LINE__);
                exit(1);
            }
            /* find the trailing newline */
            s = strchr(name, '\n');
            /* back up removing whitespace, as shortened names have ' ' added */
            while ((s != name) && isspace(*s))
                s--;
            *(s + 1) = '\0';
        }
    }
    fclose(fp);
    if (s == NULL)
    {
        fprintf(stderr, "%s:%d My process not found\n", __FILE__, __LINE__);
        exit(1);
    }
    return name;
}

static void test_set_proc_title(void)
{
    char *org_proc_title;
    char *new_title;
    char *new_proc_title;

    org_proc_title = get_process_name();
    new_title = "fubar";
    set_proc_title(new_title);
    new_proc_title = get_process_name();
    CU_ASSERT(strcmp(new_title, new_proc_title) == 0);
}


/********************************************************************
 * process_db.c tests
 ********************************************************************/
#include "../src/process_db.h"

/* 
 * Create a new process_db entry and verify that it is initialised 
 * properly. 
 */

static void test_process_db_new(void)
{
    process_h *p;
    char *w_buf = "garbage in, garbage out\n";
    char r_buf[128];

    process_db_new(&p, T_START);

    /* basic checks */
    CU_ASSERT(p->pt == T_START);
    CU_ASSERT(p->pid == 0);
    CU_ASSERT(p->r_code == -10);   /* XXX: magic number */
    CU_ASSERT(p->buffer != NULL);
    CU_ASSERT(p->buffer_pos == 0);

    /* check that our pipe works */
    CU_ASSERT(write(p->stdout[1], w_buf, strlen(w_buf)+1) == strlen(w_buf)+1);
    CU_ASSERT(read(p->stdout[0], r_buf, sizeof(r_buf)) == strlen(w_buf)+1);
    CU_ASSERT(strcmp(w_buf, r_buf) == 0);

    process_db_free(&p);
}

/* 
 * verify process_db_free() works
 */

/* see if the given integer is a file descriptor */
static int is_fd(int fd)
{
    struct stat buf;

    if (fstat(fd, &buf) == -1) {
        return 0;
    } else {
        return 1;
    }
}

static void test_process_db_free(void)
{
    process_h *p;
    int fds[2];

    /* hope we don't core dump */
    p = NULL;
    process_db_free(&p);

    /* standard free */
    process_db_new(&p, T_START);
    fds[0] = p->stdout[0];
    fds[1] = p->stdout[1];
    process_db_free(&p);

    /* check our result */
    CU_ASSERT(p == NULL);

    /* verify our pipes have closed */
    CU_ASSERT(!is_fd(fds[0]));
    CU_ASSERT(!is_fd(fds[1]));

    /* try a double-free */
    process_db_free(&p);
}

/********************************************************************
 * open_read_close.c tests
 ********************************************************************/
#include "../src/open_read_close.h"

static char *fill_file(const char *filename, int size)
{
    FILE *fp;
    char *buf;
    int i;
    int n;

    /* open file */
    fp = fopen(filename, "w");
    if (fp == NULL) {
        fprintf(stderr, "%s:%d Error opening \"%s\" for writing\n",
                __FILE__, __LINE__, filename);
        exit(1);
    }

    /* build our output buffer, filled with random characters */
    buf = (char *)malloc(size+1);
    if (buf == NULL) {
        fprintf(stderr, "%s:%d Out of memory\n", __FILE__, __LINE__);
        exit(1);
    }
    srand48(time(NULL));
    i = 0;
    while (i < size) {
        n = lrand48() & 0xFF;
        if (isprint(n)) {
            buf[i++] = n;
        }
    }
    buf[size] = '\0';

    /* output until we're done */
    if (fwrite(buf, 1, size, fp) != size) {
        fprintf(stderr, "%s:%d Error writing file\n", __FILE__, __LINE__);
        exit(1);
    }
    
    /* close */
    fclose(fp);

    /* return our random buffer */
    return buf;
}

/* 
 * Missing tests for failure of:
 *
 * - fstat() return error
 * - read() return error
 * - read() not get enough bytes
 * - close() return error
 */
static void test_open_read_close(void)
{
    char *buf;
    char *cmp_buf;

    /* verify that the read fails if we give it a bad filename */
    buf = "";
    CU_ASSERT(open_read_close("no-such-filename", &buf) == FALSE);
    CU_ASSERT(buf == NULL);

    /* XXX: how to get fstat() to fail? */

    /* try with an empty file */
    cmp_buf = fill_file("test_initng.tmp", 0);
    CU_ASSERT(open_read_close("test_initng.tmp", &buf) == TRUE);
    CU_ASSERT(buf != NULL);
    CU_ASSERT(strlen(buf) == 0);
    free(cmp_buf);

    /* try with a non-empty file */
    cmp_buf = fill_file("test_initng.tmp", 100);
    CU_ASSERT(open_read_close("test_initng.tmp", &buf) == TRUE);
    CU_ASSERT(buf != NULL);
    CU_ASSERT(strlen(buf) == 100);
    CU_ASSERT(strcmp(buf, cmp_buf) == 0);
    free(cmp_buf);

    /* try with a big file */
    cmp_buf = fill_file("test_initng.tmp", 100000);
    CU_ASSERT(open_read_close("test_initng.tmp", &buf) == TRUE);
    CU_ASSERT(buf != NULL);
    CU_ASSERT(strlen(buf) == 100000);
    CU_ASSERT(strcmp(buf, cmp_buf) == 0);
    free(cmp_buf);

    /* cleanup */
    if (remove("test_initng.tmp") != 0) {
        fprintf(stderr, "%s:%d Error %d removing file; %s\n", 
                __FILE__, __LINE__, errno, strerror(errno));
        exit(1);
    }
}

#include "../src/initng_global.h"

int main(int argc, char *argv[], char *env[])
{
    CU_pSuite suite;

    initng_new(argc, argv, env);

    CU_initialize_registry();

    suite = CU_add_suite("initng_toolbox", empty_init, empty_cleanup);
    CU_add_test(suite, "initng_calloc", test_initng_calloc);
    CU_add_test(suite, "initng_realloc", test_initng_realloc);
    CU_add_test(suite, "initng_i_strdup", test_i_strdup);
    CU_add_test(suite, "initng_i_strndup", test_i_strndup);
    CU_add_test(suite, "out of memory", test_out_of_memory);
    CU_add_test(suite, "set_proc_title", test_set_proc_title);

    suite = CU_add_suite("initng_processdb", empty_init, empty_cleanup);
    CU_add_test(suite, "process_db_new", test_process_db_new);
    CU_add_test(suite, "process_db_free", test_process_db_free);

    suite = CU_add_suite("open_read_close", empty_init, empty_cleanup);
    CU_add_test(suite, "open_read_close", test_open_read_close);

    CU_basic_run_tests();

    return 0;
}
